/*
 *
 * TODO
 */
#include <gtk/gtk.h>
#include <gtk/gtkimcontext.h>
#include <gtk/gtkimmodule.h>
#include <gtk/gtkimcontextsimple.h>
#include <gdk/gdkkeysyms.h>
#include <iconv.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "rkconv.h"
#include "rkmap.h"
#include "skk-dic.h"
#include "skk-key.h"

#define SKK_DIC_FILE "/usr/share/skk/SKK-JISYO.L"

/* functions */
/* class management */
static void im_skk_class_init (GtkIMContextClass *class);
static void im_skk_finalize(GObject *obj);
static void im_skk_init (GtkIMContext *context);

/* exported module interface */
void im_module_list(const GtkIMContextInfo ***contexts,
		    int *n_contexts);
void im_module_init(GTypeModule *module);
void im_module_exit();
GtkIMContext * im_module_create (const gchar *context_id);

static char *ascii_wide_tab[]={
    0,"","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","",
};

enum skk_mode {
  STATE_DIRECT,
  STATE_KANJI,
  STATE_CONVERTING,
  STATE_OKURI,
  STATE_LATIN,
  STATE_WIDE_LATIN,
};

typedef struct _IMSkkContext
{
  struct _GtkIMContext parent;
  struct _GtkIMContext *slave;
  enum skk_mode mode;
  struct rk_conv_context *rkc;
  int kana;
  char *head;
  char *okuri;
  char okuri_char;
  char *result;
  int cand_index;
  char *candidate;
} IMSkkContext;

typedef struct _IMContextSkkClass 
{
  GtkIMContextClass parent_class;
} IMContextSkkClass;


static GType type_im_skk;
static GObjectClass *parent_class;

static const GtkIMContextInfo im_skk_info = {
  "skk",
  "Skk(Japanese Conversion System)",
  "gtk+",
  "",
  "ja"
};

static const GtkIMContextInfo *info_list[] = {
  &im_skk_info
};

static const GTypeInfo object_info =
  {
    sizeof(IMContextSkkClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) im_skk_class_init,
    NULL,
    NULL,
    sizeof(IMSkkContext),
    0,
    (GtkObjectInitFunc) im_skk_init,
  };

#define IM_SKK_CONTEXT(obj) (GTK_CHECK_CAST((obj),type_im_skk,IMSkkContext))

/*
 * global variables for immodule
 */
static guint16 gtk_compose_ignore[] = {
  GDK_Shift_L,
  GDK_Shift_R,
  GDK_Control_L,
  GDK_Control_R,
  GDK_Caps_Lock,
  GDK_Shift_Lock,
  GDK_Meta_L,
  GDK_Meta_R,
  GDK_Alt_L,
  GDK_Alt_R,
  GDK_Super_L,
  GDK_Super_R,
  GDK_Hyper_L,
  GDK_Hyper_R,
  GDK_Mode_switch,
  0
};

static struct {
  GtkWidget *win;
  GtkWidget *menu;
} status_bar;

static IMSkkContext *current_context;

struct rk_map *hira_map;
struct rk_map *kata_map;
static struct dic_info *di;

/* input and event handling */
static void im_skk_set_client_window(GtkIMContext *, GdkWindow *);
static void im_skk_get_preedit_string(GtkIMContext *, gchar **str,
				      PangoAttrList **attrs,
				      gint *cursor_pos);
static gboolean im_skk_filter_keypress(GtkIMContext *context,
				       GdkEventKey *key);
static void im_skk_focus_in(GtkIMContext *);
static void im_skk_focus_out(GtkIMContext *);
static void im_skk_commit_cb(GtkIMContext *ic,
			     const gchar *str,
			     IMSkkContext *ac);
static void im_skk_reset(GtkIMContext *);


static void do_commit(IMSkkContext *, char *);
static struct CandidateWin *create_candidate_win();
static void release_candidate_win(struct CandidateWin *win);
static char *euc_to_utf8(char *);
static void add_input_mode(GtkWidget *, char *, int);
static void im_skk_off(IMSkkContext *);
/* skk */

static char *
euc_to_utf8(char *str)
{
  iconv_t conv;
  int len = strlen(str);
  int buflen = len * 6+3;
  char *realbuf = alloca(buflen);
  char *outbuf = realbuf;
  char *inbuf = str;
  bzero(realbuf, buflen);
  conv = iconv_open("UTF-8", "EUC-JP");
  if (conv == (iconv_t) -1) {
    return NULL;
  }
  iconv(conv,&inbuf,&len,&outbuf,&buflen);
  iconv_close(conv);
  return strdup(realbuf);
}

static int
backspace_str(char *str)
{
  int len = strlen(str);
  if (!len) {
    return -1;
  }
  if (!(str[len-1] & 0x80)) {
    str[len-1] = 0;
    return len-1;
  }
  if (len == 1) {
    str[0] = 0;
    return 0;
  }
  str[len-2] = 0;
  return len -2;
}

static void
im_skk_set_client_window(GtkIMContext *ic, GdkWindow *w)
{
  IMSkkContext *ac = IM_SKK_CONTEXT(ic);
}

static void
im_skk_commit_cb(GtkIMContext *ic,
		   const gchar *str,
		   IMSkkContext *ac)
{
  g_signal_emit_by_name(ac, "commit", str);
}

static void
update_preedit(IMSkkContext *sc)
{
  g_signal_emit_by_name(sc, "preedit_changed");
}

static char *
append_segment(PangoAttrList *attrs, char *src,
	       char *dst, int rev)
{
  PangoAttribute *attr;
  if (attrs) {
    attr = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
    attr->start_index = strlen(src);
    attr->end_index = attr->start_index + strlen(dst);
    pango_attr_list_change(attrs, attr);
  }
  src = realloc(src, strlen(src) + strlen(dst)+1);
  strcat(src, dst);
  free(dst);
  return src;
}

static char *
get_preedit_string(IMSkkContext *sc, PangoAttrList *attr)
{
  char *str = strdup("");
  char *head;
  char buf[10];
  if (sc->mode != STATE_DIRECT &&
      sc->mode != STATE_LATIN &&
      sc->mode != STATE_WIDE_LATIN) {
    char nabla[] = {0xe2, 0x96,0xbd,0};
    str = append_segment(attr, str, strdup(nabla), 0);
  }
  if (sc->candidate) {
    head = euc_to_utf8(sc->candidate);
  } else {
    head = euc_to_utf8(sc->head);
  }
  str = append_segment(attr, str, head, 0);
  if (strlen(sc->okuri)) {
    str = append_segment(attr, str, euc_to_utf8(sc->okuri), 0);
  }
  rk_get_pending_str(sc->rkc, buf, 10);
  if (strlen(buf)) {
    str = append_segment(attr, str, euc_to_utf8(buf), 0);
  }
  return str;
}

static void
im_skk_get_preedit_string(GtkIMContext *ic, gchar **str,
			  PangoAttrList **attrs,
			  gint *cursor_pos)
{
  IMSkkContext *sc = IM_SKK_CONTEXT(ic);
  char **tmp, *tmp1;
  
  if (attrs) {
    *attrs = pango_attr_list_new();
  }
  if (cursor_pos) {
    *cursor_pos = 0;
  }
  if (str) {
    tmp = str;
  } else {
    tmp = &tmp1;
  }
  if (attrs && *attrs) {
    *tmp = get_preedit_string(sc, *attrs);
  } else {
    *tmp = get_preedit_string(sc, NULL);
  }
  if (!str) {
    free(*tmp);
  }
}

static void
skk_flush(IMSkkContext *sc)
{
  rk_flush(sc->rkc);
  sc->head[0] = 0;
  sc->okuri[0] = 0;
  sc->okuri_char = 0;
  sc->candidate = 0;
  sc->mode = STATE_DIRECT;
}

static void
do_commit(IMSkkContext *ic, char *str)
{
  char *realbuf;
  realbuf = euc_to_utf8(str);
  g_signal_emit_by_name(ic, "commit", realbuf);
  free(realbuf);
}

static void
im_skk_reset(GtkIMContext *c)
{
}

static void
im_skk_focus_in(GtkIMContext *c)
{
  current_context = IM_SKK_CONTEXT(c);
}

static void
im_skk_focus_out(GtkIMContext *c)
{
}

static void
im_skk_off(IMSkkContext *ac)
{
  ac->mode = STATE_LATIN;
}

static void
update_candidate(IMSkkContext *sc)
{
  if (sc->candidate) {
    free(sc->candidate);
  }
  sc->candidate = nth_candidate(sc->result, sc->cand_index);
}

static void
begin_conversion(IMSkkContext *sc)
{
  sc->result = skk_search(di, sc->head, sc->okuri_char);
  if (!sc->result) {
    skk_flush(sc);
    return ;
  }
  sc->cand_index = 0;
  update_candidate(sc);
  sc->mode = STATE_CONVERTING;
}

static void
proc_state_converting(IMSkkContext *sc, GdkEventKey *key)
{
  if (skk_cancel(key)) {
    strcat(sc->head, sc->okuri);
    sc->okuri[0] = 0;
    if (sc->candidate) {
      free(sc->candidate);
      sc->candidate = 0;
    }
    sc->mode = STATE_KANJI;
    return ;
  }
  if (skk_next_candidate(key)) {
    sc->cand_index ++;
    update_candidate(sc);
    if (!sc->candidate) {
      sc->cand_index = 0;
      update_candidate(sc);
    }
    return ;
  }
  do_commit(sc, sc->candidate);
  if (strlen(sc->okuri)) {
    do_commit(sc, sc->okuri);
  }
  skk_flush(sc);
  if (!skk_commit(key)) {
    rk_push_key(sc->rkc, key->keyval);
  }
}

static void
proc_state_okuri(IMSkkContext *sc, GdkEventKey *key)
{
  char buf[10];
  if (skk_cancel(key)) {
    skk_flush(sc);
    return ;
  }
  rk_push_key(sc->rkc, key->keyval);
  rk_result(sc->rkc, buf, 10);
  strcat(sc->okuri, buf);
  if (strlen(sc->okuri)) {
    begin_conversion(sc);
    return ;
  }
}

static void
proc_state_kanji(IMSkkContext *sc, GdkEventKey *key)
{
  char buf[10];
  int kv = key->keyval;

  if (skk_begin_conv(key)) {
    begin_conversion(sc);
    return ;
  }
  if (skk_commit(key)) {
    do_commit(sc, sc->head);
    skk_flush(sc);
    return ;
  }
  if (skk_cancel(key)) {
    skk_flush(sc);
    return ;
  }
  if (skk_backspace(key)) {
    int res;
    res = backspace_str(sc->head);
    if (res <= 0) {
      skk_flush(sc);
    }
    return ;
  }
  if (key->state & GDK_SHIFT_MASK) {
    kv = tolower(kv);
    if (strlen(sc->head)) {
      sc->mode = STATE_OKURI;
      sc->okuri[0] = 0;
      sc->okuri_char = kv;
    }
  }

  rk_push_key(sc->rkc, kv);
  rk_result(sc->rkc, buf, 10);
  if (sc->mode == STATE_OKURI) {
    strcat(sc->okuri, buf);
    if (strlen(sc->okuri)) {
      begin_conversion(sc);
      return ;
    }
    return ;
  } else {
    strcat(sc->head, buf);
  }
}

static void
proc_state_direct(IMSkkContext *sc, GdkEventKey *key)
{
  char buf[10];
  int kv = key->keyval;

  if (skk_latin(key)) {
    sc->mode = STATE_LATIN;
    return ;
  }
  if (skk_wide_latin(key)) {
    sc->mode = STATE_WIDE_LATIN;
    return ;
  }
  if (skk_kana_toggle(key)) {
    sc->kana = !sc->kana;
    if (sc->kana) {
      rk_select_map(sc->rkc, kata_map);
    } else {
      rk_select_map(sc->rkc, hira_map);
    }
    skk_flush(sc);
    return ;
  }
  if ((key->state & GDK_SHIFT_MASK) &&
      kv < 128 &&
      isalpha(kv)) {
    kv = tolower(kv);
    sc->mode = STATE_KANJI;
    sc->head[0] = 0;
  }
  if (skk_cancel(key)) {
    rk_flush(sc->rkc);
    return ;
  }

  if (kv < 128 && !isalpha(kv)) {
    char buf[2];
    buf[0] = kv;
    buf[1] = 0;
    do_commit(sc, buf);
    return ;
  }
  rk_push_key(sc->rkc, kv);
  rk_result(sc->rkc, buf, 10);
  if (sc->mode == STATE_DIRECT) {
    if (strlen(buf)) {
      do_commit(sc, buf);
    }
  } else {
    strcat(sc->head, buf);
  }
}

static gboolean
proc_skk_keypress(IMSkkContext *sc, GdkEventKey *key)
{
  switch (sc->mode) {
  case STATE_DIRECT:
    printf("Direct\n");
    proc_state_direct(sc, key);
    break;
  case STATE_KANJI:
    printf("Kanji\n");
    proc_state_kanji(sc, key);
    break;
  case STATE_CONVERTING:
    printf("Conversion\n");
    proc_state_converting(sc, key);
    break;
  case STATE_OKURI:
    printf("Okuri\n");
    proc_state_okuri(sc, key);
    break;
  }
  update_preedit(sc);
  return FALSE;
}

static gboolean
im_skk_filter_keypress(GtkIMContext *context,
		       GdkEventKey *key)
{
  IMSkkContext *sc = IM_SKK_CONTEXT(context);
  int i;
  if (key->type == GDK_KEY_RELEASE) {
    return FALSE;
  }
  for (i = 0; gtk_compose_ignore[i]; i++) {
    if(key->keyval == gtk_compose_ignore[i]) {
      return FALSE;
    }
  }
  if (sc->mode == STATE_LATIN) {
    if (skk_on_key(key)) {
      sc->mode = STATE_DIRECT;
      return FALSE;
    }
    return gtk_im_context_filter_keypress(sc->slave, key);
  }
  if (sc->mode == STATE_WIDE_LATIN) {
    int kv;
    if (skk_on_key(key)) {
      sc->mode = STATE_DIRECT;
      return FALSE;
    }
    kv = key->keyval;
    if (kv < 127 && kv > 32) {
      do_commit(sc, ascii_wide_tab[kv-32]);
    }
    return FALSE;
  }

  return proc_skk_keypress(sc, key);
}

/*
 * class init
 */
void
im_skk_class_init (GtkIMContextClass *class)
{
  GtkIMContextClass *im_context_class =
    GTK_IM_CONTEXT_CLASS(class);
  GObjectClass *object_class = G_OBJECT_CLASS(class);

  parent_class = g_type_class_peek_parent (class);

  im_context_class->set_client_window = im_skk_set_client_window;
  im_context_class->filter_keypress = im_skk_filter_keypress;
  im_context_class->reset = im_skk_reset;
  im_context_class->get_preedit_string = im_skk_get_preedit_string;
  im_context_class->focus_in = im_skk_focus_in;
  im_context_class->focus_out = im_skk_focus_out;
  object_class->finalize = im_skk_finalize;
}

/*
 * The Constructor
 */
void
im_skk_init (GtkIMContext *context)
{
  IMSkkContext *sc = IM_SKK_CONTEXT(context);
  sc->slave = g_object_new(GTK_TYPE_IM_CONTEXT_SIMPLE, NULL);
  g_signal_connect(sc->slave, "commit",
		   G_CALLBACK(im_skk_commit_cb), sc);
  sc->mode = STATE_LATIN;
  sc->rkc = rk_context_create();
  sc->kana = 0;
  rk_select_map(sc->rkc, hira_map);
  sc->head = (char *)malloc(100);
  sc->okuri = (char *)malloc(100);
  skk_flush(sc);

}

/*
 * The Destructor
 */
void
im_skk_finalize(GObject *obj)
{
  IMSkkContext *ac = IM_SKK_CONTEXT(obj);
  g_signal_handlers_disconnect_by_func(ac->slave,
				       im_skk_commit_cb,
				       ac);
  free(ac->head);
  g_object_unref(ac->slave);
  parent_class->finalize (obj);
}

GtkIMContext *
im_module_create (const gchar *context_id)
{
  if (strcmp (context_id, "skk") == 0) {
    GtkObject *obj = g_object_new(type_im_skk, NULL);
    return GTK_IM_CONTEXT (obj);
  }
  return NULL;
}

/*
 *
 */
void
im_module_list(const GtkIMContextInfo ***contexts,
	       int *n_contexts)
{
  *contexts = info_list;
  *n_contexts = G_N_ELEMENTS(info_list);
}

void
im_module_init(GTypeModule *type_module)
{
  const char *fn = SKK_DIC_FILE;
  hira_map = rk_map_create(rk_rule_hiragana);
  kata_map = rk_map_create(rk_rule_katakana);
  di = open_dic(fn);
  if (!di) {
    printf("Failed to open %s\n",fn);
    return ;
  }

  type_im_skk =
    g_type_module_register_type(type_module,
				GTK_TYPE_IM_CONTEXT,
				"GtkIMContextSkk",
				&object_info, 0);
}

void
im_module_exit()
{
}
